/*
  This program is part of Attribute Zone and is in the public domain.

  The function of this program is to read the uncompressed levels (RAM images),
  to make Huffman analysis of them, and to write out the compressed data and a
  assembly code program to decompress them.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define new(x) malloc(sizeof(x))

typedef struct Tree {
  struct Tree*zero;
  struct Tree*one;
  unsigned char data;
} Tree;

typedef struct {
  char data[64];
} Encoding;

static unsigned long hitcount[64];
static Tree*trees[64];
static Tree*root;
static char rle_usage[32];

static Encoding enc_tiles[32];
static Encoding enc_sprites[32];
static Encoding enc_rle[32];
static Encoding enc_xnextobj[32];
static Encoding enc_ynextobj[32];

static int split_point=255;
static int lvladdr[256];
static int lvlsize[256];

static int bitcount;
static int filepos;
static int bitval;
static int fakepos;

static void write_bit(FILE*fp,int b) {
  if(bitcount==8) {
    if(fp) {
      fputc(bitval,fp);
      filepos++;
    } else {
      fakepos++;
    }
    bitval=bitcount=0;
  }
  bitcount++;
  bitval<<=1;
  bitval|=b&1;
}

static void final_bit(FILE*fp) {
  while(bitcount!=1) write_bit(fp,bitcount);
  bitval=bitcount=0;
}

static inline void write_bitstr(FILE*fp,char*s) {
  while(*s) write_bit(fp,*s++);
}

static inline void write_huffstr(FILE*fp,Encoding*enc,int idx) {
  write_bitstr(fp,enc[idx].data);
}

static inline void convert_trees(Tree*r,char*s,Encoding*enc) {
  char n[64];
  if(!r) return;
  if(r->zero || r->one) {
    strcpy(n,s);
    n[strlen(s)+1]=0;
    n[strlen(s)]='0';
    convert_trees(r->zero,n,enc);
    n[strlen(s)]='1';
    convert_trees(r->one,n,enc);
  } else if(r->data<32) {
    strcpy(enc[r->data].data,s);
  }
}

static void huffman_analysis(Encoding*enc) {
  int i,m0,m1;
  int h=32;
  for(i=0;i<64;i++) {
    trees[i]=new(Tree);
    trees[i]->zero=trees[i]->one=0;
    trees[i]->data=i;
  }
  for(i=32;i<64;i++) hitcount[i]=0;
  root=trees[32];
  for(;;) {
    m0=m1=-1;
    for(i=0;i<64;i++) {
      if(hitcount[i] && (m0==-1 || hitcount[i]<hitcount[m0])) m1=m0,m0=i;
      else if(hitcount[i] && (m1==-1 || hitcount[i]<hitcount[m1])) m1=i;
    }
    if(m0==-1 || m1==-1) break;
    hitcount[h]=hitcount[m0]+hitcount[m1];
    hitcount[m0]=hitcount[m1]=0;
    trees[h]->zero=trees[m0];
    trees[h]->one=trees[m1];
    root=trees[h];
    ++h;
  }
  convert_trees(root,"",enc);
  for(i=0;i<64;i++) {
    hitcount[i]=0;
    free(trees[i]);
  }
}

static inline void read_rle_usage(void) {
  FILE*fp=fopen("rle.dat","r");
  int i;
  while(!feof(fp)) {
    fscanf(fp,"%*[ \r\n\t]");
    if(fscanf(fp,"%i",&i) && i>=0 && i<32) rle_usage[i]=1;
  }
  fclose(fp);
}

static inline void analyze_levels(void) {
  char n[32];
  int v,i,j,k,x,y,z;
  FILE*fp;
  for(v=0;v<256;v++) {
    sprintf(n,"level/%d",v);
    fp=fopen(n,"rb");
    if(!fp) break;
    fseek(fp,0x1E,SEEK_SET);
    j=fgetc(fp);
    i=k=-1;
    for(y=0;y<26;y++) {
      fseek(fp,0x400+(y<<5)+j,SEEK_SET);
      for(x=0;x<30;x++) {
        z=fgetc(fp)&0x1F;
        if(i!=z || !k--) {
          if(rle_usage[z]) i=z,k=32;
          hitcount[z]++;
        }
      }
    }
    fclose(fp);
  }
  huffman_analysis(enc_tiles);
  for(v=0;v<256;v++) {
    sprintf(n,"level/%d",v);
    fp=fopen(n,"rb");
    if(!fp) break;
    fseek(fp,0x204,SEEK_SET);
    for(i=1;i<64;i++) {
      y=fgetc(fp);
      z=fgetc(fp);
      fgetc(fp);
      x=fgetc(fp);
      if(y<0xF0) hitcount[z]++;
    }
    fclose(fp);
  }
  huffman_analysis(enc_sprites);
  for(v=0;v<256;v++) {
    sprintf(n,"level/%d",v);
    fp=fopen(n,"rb");
    if(!fp) break;
    fseek(fp,0x204,SEEK_SET);
    j=0x0F;
    for(i=1;i<64;i++) {
      y=fgetc(fp);
      z=fgetc(fp);
      fgetc(fp);
      x=fgetc(fp);
      if(y<0xF0) {
        hitcount[(y-j)>>3]++;
        j=y;
      }
    }
    fclose(fp);
  }
  hitcount[(0xD7-j)>>3]++;
  huffman_analysis(enc_ynextobj);
  for(v=0;v<256;v++) {
    sprintf(n,"level/%d",v);
    fp=fopen(n,"rb");
    if(!fp) break;
    fseek(fp,0x204,SEEK_SET);
    for(i=1;i<64;i++) {
      y=fgetc(fp);
      z=fgetc(fp);
      fgetc(fp);
      x=fgetc(fp);
      if(y<0xF0) hitcount[(x-0x08)>>3]++;
    }
    fclose(fp);
  }
  huffman_analysis(enc_xnextobj);
  for(v=0;v<256;v++) {
    sprintf(n,"level/%d",v);
    fp=fopen(n,"rb");
    if(!fp) break;
    fseek(fp,0x1E,SEEK_SET);
    j=fgetc(fp);
    i=k=-1;
    for(y=0;y<26;y++) {
      fseek(fp,0x400+(y<<5)+j,SEEK_SET);
      for(x=0;x<30;x++) {
        z=fgetc(fp)&0x1F;
        if(z!=i || k==31) {
          if(i!=-1) hitcount[k]++;
          i=-1;
          if(rle_usage[z]) i=z,k=-1;
        }
        k++;
      }
    }
    if(i!=-1) hitcount[k]++;
    fclose(fp);
  }
  huffman_analysis(enc_rle);
}

#define color_addr ((((((par>>5)+y)<<1)&0xF8)|0xC0)+(((par&1)+x)>>2))
#define color_data (3<<((x&2)|((y&2)<<1)))

static void parse_one_level(FILE*in,FILE*out) {
  int par,i,x,y,z,rc,rt;
  unsigned char at[64];
  memset(at,0,64);
  // Parity
  fseek(in,0x01E,SEEK_SET);
  par=fgetc(in);
  write_bit(out,par>>5);
  write_bit(out,par);  
  // Tiles
  rc=rt=-1;
  for(y=0;y<26;y++) {
    fseek(in,0x400+(y<<5)+par,SEEK_SET);
    for(x=0;x<30;x++) {
      z=fgetc(in)&0x1F;
      if((z!=rt || rc==31) && rt!=-1) {
        write_huffstr(out,enc_rle,rc);
        rc=rt=-1;
      }
      if(rt!=-1) {
        rc++;
      } else if(rle_usage[z]) {
        rc=0;
        rt=z;
        write_huffstr(out,enc_tiles,z);
      } else {
        write_huffstr(out,enc_tiles,z);
      }
      if(z&16) at[color_addr&63]|=color_data;
    }
  }
  if(rt!=-1) {
    for(i=rc;i<32;i++) if(enc_rle[i].data[0] && !enc_rle[rc].data[strlen(enc_rle[i].data)]) rc=i;
    write_huffstr(out,enc_rle,rc);
  }
  // Colors
  fseek(in,0x7C0,SEEK_SET);
  for(i=0;i<64;i++) {
    z=fgetc(in);
    while(at[i]) {
      if(at[i]&1) write_bit(out,z);
      at[i]>>=1;
      z>>=1;
    }
  }
  // Sprites
  y=0x0F;
  fseek(in,0x204,SEEK_SET);
  for(i=1;i<64;i++) {
    z=fgetc(in);
    if(z==0xFF) break;
    write_huffstr(out,enc_ynextobj,(z-y)>>3);
    y=z;
    rt=fgetc(in);
    fgetc(in);
    x=(fgetc(in)>>3)-1;
    write_huffstr(out,enc_xnextobj,x);
    write_huffstr(out,enc_sprites,rt);
  }
  write_huffstr(out,enc_ynextobj,26-(y>>3));
  // Finish
  final_bit(out);
}

static inline void print_attribute_grid(int par) {
  int x,y;
  for(y=0;y<26;y++) {
    for(x=0;x<30;x++) printf("%02X",color_addr);
    puts("");
    for(x=0;x<30;x++) printf(" %1x", ((x&2)|((y&2)<<1)) );
    puts("");
  }
}

static inline void parse_levels(void) {
  char n[32];
  int x,y;
  FILE*fp;
  FILE*ofp=fopen("lvldata.0","wb");
  y=0;
  filepos=fakepos=0x0C00;
  for(x=0;x<256;x++) {
    sprintf(n,"level/%d",x);
    fp=fopen(n,"rb");
    if(!fp) break;
    parse_one_level(fp,0);
    if(fakepos>=0x2000) {
      fakepos-=filepos;
      while(filepos++<0x2000) fputc(42,ofp);
      fclose(ofp);
      sprintf(n,"lvldata.%d",++y);
      ofp=fopen(n,"wb");
      filepos=0;
      if(y==8) split_point=x;
    }
    lvlsize[x]=fakepos-filepos;
    fakepos=filepos;
    lvladdr[x]=(y<<13)|filepos;
    parse_one_level(fp,ofp);
    fakepos=filepos;
    fclose(fp);
  }
  while(filepos++<0x2000) fputc(42,ofp);
  fclose(ofp);
}

static inline void write_levelptr_inc(void) {
  FILE*fp=fopen("levelptr.inc","w");
  int i;
  fputs("; This file is automatically generated and should not be adjusted manually.\n\nlvlptrl",fp);
  for(i=0;i<256;i++) fprintf(fp,"\tdb %d ; Level %d: %d bytes\n",lvladdr[i]&0xFF,i,lvlsize[i]);
  fputs("lvlptrh",fp);
  for(i=0;i<256;i++) fprintf(fp,"\tdb %d\n",(lvladdr[i]>>8)&0xFF);
  fprintf(fp,"lvlsplt\t= %d\n",split_point);
  fclose(fp);
}

static void write_decoder(FILE*fp,Encoding*enc,char*seq,int n,int*cp,const char*rlbl,const char*lbl,const char*fstr) {
  int i,e;
  int v=0;
  for(i=0;i<32;i++) {
    if(enc[i].data[0] && !memcmp(seq,enc[i].data,n)) {
      switch(enc[i].data[n]|v) {
        case 0x00:
          fprintf(fp,fstr,i,rle_usage[i]?rlbl:lbl);
          return;
        case 0x30:
        case 0x31:
          fprintf(fp,"jsr getbit\n\tbc%c e_%d\n\t",enc[i].data[n]&1?'c':'s',e=++*cp);
          v=enc[i].data[n]<<1;
          write_decoder(fp,enc,enc[i].data,n+1,cp,rlbl,lbl,fstr);
          fprintf(fp,"e_%d\t",e);
          break;
        case 0x71:
        case 0x72:
          write_decoder(fp,enc,enc[i].data,n+1,cp,rlbl,lbl,fstr);
          return;
      }
    }
  }
}

static inline void write_leveldec_inc(void) {
  FILE*fp=fopen("leveldec.inc","w");
  int i;
  // Description
  fputs("; This file is automatically generated and should not be adjusted manually.\n\n\t;TILES:\n",fp);
  for(i=0;i<32;i++) fprintf(fp,"\t;  $%02X : %s\n",i,enc_tiles[i].data);
  fputs("\t;SPRITES:\n",fp);
  for(i=0;i<16;i++) fprintf(fp,"\t;   $%01X : %s\n",i,enc_sprites[i].data);
  fputs("\t;RLE:\n",fp);
  for(i=0;i<32;i++) fprintf(fp,"\t;   %2d : %s\n",i+1,enc_rle[i].data);
  fputs("\t;XNEXTOBJ:\n",fp);
  for(i=0;i<30;i++) fprintf(fp,"\t;   %2d : %s\n",i,enc_xnextobj[i].data);
  fputs("\t;YNEXTOBJ:\n",fp);
  for(i=0;i<26;i++) fprintf(fp,"\t;   %2d : %s\n",i,enc_ynextobj[i].data);
  fputs("\t;RLE_USAGE:\n",fp);
  for(i=0;i<32;i++) if(rle_usage[i]) fprintf(fp,"\t;  $%02X\n",i);
  // Tile decoder
  fputs("\ndectil\t",fp);
  i=0;
  write_decoder(fp,enc_tiles,"",0,&i,"rle","tilp","lda #$%02X\n\tjmp dec%s\n");
  // RLE decoder
  fputs("\ndecrle\tpha\n\t",fp);
  write_decoder(fp,enc_rle,"",0,&i,"","","lda #%d+1\n\tjmp dectilr%s\n");
  // Y position decoder
  fputs("\ndecyp\t",fp);
  write_decoder(fp,enc_ynextobj,"",0,&i,"","","lda #%d*8\n\tjmp decxp%s\n");
  // X position decoder
  fputs("\n\torg *-3\ndecxp\tclc\n\tadc <t0\n\tsta <t0\n\t",fp);
  write_decoder(fp,enc_xnextobj,"",0,&i,"","","lda #%d*8+8\n\tjmp decspr%s\n");
  // Sprite decoder
  fputs("\n\torg *-3\ndecspr\tsta $203,y\n\t",fp);
  write_decoder(fp,enc_sprites,"",0,&i,"","","lda #$%01X\n\tjmp decsprp%s\n");
  // Finished
  fclose(fp);
}

int main(int argc,char**argv) {
  read_rle_usage();
  analyze_levels();
  parse_levels();
  write_levelptr_inc();
  write_leveldec_inc();
  return 0;
}
